( 3iBeats-menu.fs -- Assad Ebrahim - module providing experimental menu capabilities )

\ Firmware: 3iI2C.avr (+ 3iForth.avr) for OLED menu
\ Software: see help-inputbeats
\ Hardware: 8R speaker on pin14 (A0) - with 220R resistor, limits max current draw to 17mA at 100% duty cycle 5V

\ DESIGN NOTE: How to develop menus.  
\   First develop them on the PC with PC keys and scaffolding.  Get the function right.  Keys can be monitored in fast real-time.
\   Then develop them using buttons on the uC.  First test tethered mode.  Lag is ok.  Get the function duplicated.
\   Then write the embedded loops to monitor.  Now test for usability.  No lag.  No duplicate presses.  Debounce if needed.

\ AVR dependencies: 3iOLED.avr
\ Forth dependencies: 3iOLED.fs

\ TODO: migrate from 3iOLED.avr (heavy OLED driver from Adafruit) to 3iI2C.avr (slim OLED driver, developed for TinyPhoto)

\ Hardware: Yellow-Blue OLED display 128x64 pixels,  16-top pixels are YELLOW, remaining 48 pixels are BLUE
\ DESIGN: to read hardware buttons in real-time, need to debounce AND maintain state so only 1 press gets registered...


\ OLED-INIT (3iI2C.avr, 3iI2C.fs)
\ OLED-INIT (3iOled.avr, 3iOled.fs)
\ : reset-menu oled-init 2 oled-textsz ;  \ only with ADAFRUIT driver

\ DESIGN: beats-menu is complex because it requires maintaining consistently three elements:
\         internal state-logic (for algorithm), stack (for data integrity), and display visible to user

\  : menu-text ( r c sptr slen -- ) >r >r >r 1 - 8 * r> set-pos drop r> r> ..oled ;
\  \ Ex: 7 1 - 8 * 0 set-pos drop s" 5-beat 1-rest"
\  : draw-menu ( -- ) \ Draw to OLED
\  oled-clear 50 ms
\  5 0 s" __INPUT BEAT____" menu-text 
\  6 0 s" 1beat 2rest   "   menu-text  
\  7 0 s" 3clr 4del 5CLR" menu-text
\  8 0 s" 6-play 7-exit"   menu-text  
\  0 0 set-pos drop 50 ms
\  ;

\ Design: _stop is a flag tied to a button providing a way to re-enter edit mode -- could make the exit & play button a toggle...
variable _stop
: reset-stop  0 _stop ! ;
: stop        1 _stop ! ;
: stop?       _stop @ ;

\ Design: One beat per bit, either beat (1) or rest (0).  4 beats per bar.  2 bars per 8-bit byte
\ Design: Working in NIBBLES - 4 beats to a quarternote
variable bitcount
variable bytecount  \ actually this is half-bar counts...
: bitcount@    bitcount @ ;
: bitcount!    bitcount ! ;
: resetbits    0 bitcount! ;
: bitcount++   bitcount@ 1 + bitcount! ;
: clearbits    bitcount@ dup 0 > if ndrop resetbits else drop then ;  \ clear stack and reset bits

: bytecount@   bytecount @ ;
: bytecount!   bytecount ! ;
: resetbytes   0 bytecount! clearbits ;
: bytecount++  bytecount@ 1 + bytecount! ;
: bytecount--  bytecount@ 1 - bytecount! ;
: clearbytes   clearbits bytecount@ dup 0 > if ndrop resetbytes else drop then ;   \ clear bytes and beats
: addbyte      >b4 bytecount++ resetbits ;
: playbar      bytecount@ 4 mod dup 0 = if drop 4 then Nplayonce drop ;  \ include beats.fs - plays latest bar 
: playall-loop bytecount@ Nplayloop drop ; \ play all bytes

\ DESIGN: building the beat-byte on the stack
: beatstack  bytecount@ dup 0 > if
                cr ." Beatstack: " visualize-beats drop
                cr ."     (hex): " .sh  
                cr ."  bitct : Bytct  " bitcount@ . ." , " bytecount@ . 
                cr ." Tempo: " beatdur@ . 
                cr
             else drop 
             then ;
\ DESIGN: LIGHT-weight menu building framework - 20 lines of Forth code
\ DESIGN NOTES: every user visible input needs to be checked to avoid crashes due to unexpected user input
: >beat   ( ." beat "   )   1 dup . bitcount++ ;
: >rest   ( ." rest "   )   0 dup . bitcount++ ;
: >clr  bitcount@ 0 > if ." | clearing bits " clearbits  beatstack else  ." ." then ;  \ clear beats in current byte, but previous bytes unchanged
: >del    >clr bytecount@ 0 > if ." | byte deleted " drop bytecount-- beatstack else ." ." then ;  \ delete last byte (clear bits)
: >clrall bitcount@ bytecount@ + 0 > if ." | clearing all "  cr  clearbytes beatstack else ." ." then ;
: >tempo  metronome  beatdur@ . ;
: >play   bitcount@ bytecount@ + 0 > if  >clr
      bytecount@ 0 > 
      if ." | playing " playall-loop  key drop
      else ." No beats to play!" cr 
      then 
  else ." ." then ;
: >stop   ." | stopping" stop ;

\ ==== MENU UPDATE SECTION
: >help cr  ." | Help: 0-HELP, 1-beat, 2-rest, 3-clear bit, 4-delete byte, "
        cr  ." |       5-clear all, 6-tempo, 7-play loop, 8-exit to dev" cr ;
create menu-table ' >help , ' >beat , ' >rest , ' >clr , ' >del , ' >clrall , ' >tempo , ' >play , ' >stop , 
\ any other key enters edit mode
: menu-valid-selection? ( choice -- choice t/f ) dup 0 8 between? ;
\ --------------------------

: menu-select ( -- choice ) ekey '0' - ;
: menu-action ( index -- <depends on action> ) cells menu-table + @ execute ;
: edit-mode  
    ." * Input 16th note beats *" cr
    beatstack
;
: nibble-complete? ( -- t/f ) bitcount@ 4 = ;

\ Design: user enters beat entry mode.  Once happy, beat plays.  If not happy, reset processor...  but you lose everything.
\ TODO: need a way to re-enter design/edit mode
: [input-beats]
    >help
    edit-mode begin 
          menu-select
          menu-valid-selection?
          if menu-action stop? 
             if exitloop 
             else
                nibble-complete? if addbyte playbar then
                continue 
             then 
          else drop exitloop 
          then 
       until
;
\ TODO: add space after 4 note/rests
\ TODO: scroll a line after each second 8 beats
\ TODO: show on user display

: input-beats  \ input without reset
    reset-stop  begin  [input-beats] stop?  until  ;


: beats ( -- )  \ input rhythm from PC keyboard, 1 sixteenth note beat at a time
                \ 1000 1000 1011 0000  1010 1011 0000 0000  (2 bars)
                \ 1000 1000 1011 0000  1010 1111 1000 0000  (4 bars)
    resetbytes ( draw-menu ) input-beats ;


\ TODO: Or is the right interface also a direct byte entry?
\ TODO: What about saving beats?
\ TODO: editor for beats - delete a bar at a time...

 
\ NEXT: extend to 1 byte 
\  add bytes
\  interface with SFX generator

: play-tones ( <notes> N -- ) nrev 0 do beat-tone // loop ;

: setup-beatsboard
   14 setspeaker 
   init-beats
   f_ a_ d_ a_  4 play-tones // //  d_ d_  2 play-tones // // 
   init-percustones
   \ oled-init
   \ buttons-init  TODO: merge buttons around keypad (D2-D9)
   ;


\ =======================================
: help-inputbeats
  cr
  cr ." HELP-INPUTBEATS"
  cr 
  cr ." include   3iForth.fs   3iI2C.fs              3iSound-4x4.fs    3iKeypad.fs"
  cr ." include   3iBeats.fs   3iBeats_cookbook.fs   3iBeats_menu.fs"
  cr
\  cr ." include 3iRecord.fs, 3iShift.fs, 3iPianoLights.fs, 3iSound-4x4.fs"
  cr ." RealTerm   1 open-com   rx"
  cr 
  cr ." setup-beatsboard "
  cr 
  cr ." beats (fresh)   input-beats (loads from stack)"
  cr ." 6 visualize-beats    6 Nbeats"
  cr 
  cr ." ex: $B 6 4 $F $E $C  tempo: 220"
  cr ." ex: 1000 1000 1011 0000   1010 1011    [  136 176 171  3 Nseq-msb  ] "
  cr ." 173 1 "
  cr 
  cr ." 136 176 171  0  4 bytecount!  input-beats"
;

help-inputbeats


\  ===========================
\ V1 - 2020-11-29---Sun---09:07 - started developing menu capability for OLED screen on Piano PCB - 4 keys (ESC, LFT, RGT, OK)
                                \ envisioned left/right keys to raise/lower octaves
                                \ just design concept, didn't actually build menu.  the octave movement was done in piano-pcb code
\ V2 - 2021-05-27---Thu---08:04 - reworked OLED.fs interface design to streamline it
  \ designing menu capability for 5-button Beat Sequencer on OLED display with notes/rests (beats.fs/.avr) - OLED driver is Adafruit
\ 5-28 - REBUILT: A working, rudimentary, menu for entering a 2-note beat
\ 5-29 - ADDED: byte-to-beat both LSB for shift-in registers and MSB for key-input
\ 5-30 - FIXED: added lsb for shift register use, msb for key/button entry use 
\ 5-30 - ADDED: readpot/setpot to adjust the rate of the beat! (used B10k pots).  ADDED: potentiometer rate control to beats
\ V3 - 2021-07-26---Mon---03:44 - structured design of beats menu around internal state-logic, stack integrity, and user display
     \ - finished first two and tested: 2021-07-26---Mon---05:28
\ 2021-07-27---Tue---23:29 - finished beats sequencer entry using 4x4 matrix keypad (4 beats per quarter note, entered with hex digit
\ 2021-07-28---Wed---07:31 - improved PC menu entry 16th note at a time and playback -- this is more intuitive than 4x4 matrix keypad

\ TODO: add ?/ help menu in real-time
\ TODO: extract out keypad code...
\ TODO: add external keypad...
\ TODO: add better external keypad --- one more spacing between keys, wires that don't break off (need strain-relief)
\     use a cable or pins (doubleheight)  Maybe 8 keys and a shift-register...  Modular keyboard that can extend as needed...
\     you add keys without adding pins... and can then program them for what you need.  batch of 8 keys


\ THEORY: 2021-07-28---Wed---12:23
( We have become so used to developing on operating systems that provide usable interfaces for users --- but at this cost of massive bloat, toolchain and compilation complexity.
  With low fat computing, we showcase how we can develop tight, functional, fun, useful, and easy to use systems on a tiny computing footprint, power efficient enough to run off a single cell 3V or 1.5V battery.
  Forth is a key enabler, with a high dynamic range covering low level and high level capabilities.
  Additionally, Forth is efficient to develop and test in an interactive environment, highly modular, promotes literate programming or you end up with an unreadable mess, modular design thinking with extreme refactoring, and simplicity.  
It is also quite portable with interface to hardware limited to the Forth virtual machine.  It is even possible to build a 3-instruction Forth, as I have done for the Arduino.
  In Forth, we can have a lightweight menu-building framework in 20 lines of Forth code!
)

