( * 7 key ACTIVE LOW keypad / drumpad - Assad Ebrahim )

8 constant  CALL_KEYPAD     \ poll keypad (in 3iKeypad.fs OR 3iDrumPad.fs) - in 3iKeypad.avr or 3iDrumPad.avr
26 constant CALL_LOOPMUSICKEYS  \ embedded loop polling keys and responding with tones - 3iDrumPad.avr

: drumroll ( n -- ) \ high to low
   7 0 do 6 i - [play-tone] // loop ;

E____ constant kick-drum \ E1 bass at 41 Hz, alternative B0 31 Hz (lowest Arduino generated tone)
E__   constant low-tom   \ E3 at 165 Hz
A__   constant mid-tom   \ A3 at 220 Hz (fourth above E3)
B__   constant high-tom  \ B3 at 247 Hz (fifth above E3)
E_    constant snare     \ E4 at 330 Hz 
B_    constant hi-hat    \ B4 at 494 Hz
: tones-drum-kit  \ tuned in key of E in fourths (A) / fifths (B)
     kick-drum
     B___       
     low-tom
     mid-tom
     high-tom
     snare 
     hi-hat    7 ;
: tones-black-key-groove  e%_ g%_ a%_ b_    4 ;
: tones-tablaCG   ( -- <tones> n ) g___  c'  c' c' 4 ;   \ for rasra3 dum-bateta  C=dayan, G=bayan dugga
: tones-tablaGC   ( -- <tones> n ) c___  g'  g' g' 4 ;   \ for rasra3 dum-bateta  C=dayan, G=bayan dugga
: tones-tablaDA   ( -- <tones> n ) a___  d_  d_ d_ 4 ;   \ for rasra3 dum-bateta
: tones-twohanddrums
     kick-drum
     kick-drum
     low-tom
     low-tom
     mid-tom
     snare
     snare
     7 ;
     

: setreg-tones ( n -- )
    0 do 
        i cells percustones + @ freq i setreg
      loop 
;


\ : keypad-init13       \ set PIANO 13 direct keys to INPUT_PULLUP
\     13 2 do i pullup loop  \ pins 2 through 12 inclusive
\     14 pullup              \ A0
\     15 pullup              \ A1
\     100 ms ; 

: keypad-init7       \ set 7 key keypad with direct button to pin connections to INPUT_PULLUP (NC on 5V, NO on 0V)
    10 3 do i pullup loop 
    100 ms ;
: poll-keypad-slow  ( -- ) \ purely Forth based software (slow)
    begin 10 3 do i pinread not if 
        i . cr   \ CODE REVIEW - can make this response word
        i 3 - [play-tone]
    then loop key? until key drop ;

\ TODO: encapsulate this into drum-pad design, repeated hold means repeated rapid strikes instead of solid tone sustain
\ TODO: what happens when multiple simult keys pressed?

create keypad-chars '0' c, '1' c, '2' c, '3' c, '4' c, '5' c, '6' c, '7' c, '8' c, '9' c, 'A' c, 'B' c, 'C' c, 'D' c, '#' c, '*' c,
create keypad-nums  0 c, 1 c, 2 c, 3 c, 4 c, 5 c, 6 c, 7 c, 8 c, 9 c, 10 c, 11 c, 12 c, 13 c, 14 c, 15 c,


\ ---------------------------------------
\ FORTH CONTROLLED KEYPAD LOOPING

\ DESIGN: key press flag signals through the layers of the loop when to stop polling
variable pollstate
: pollstate! ( n -- ) pollstate ! ;
: pollstate@ ( -- n ) pollstate @ ;
: pollcontinue continue pollstate! ;
: pollstop     exitloop pollstate! ;


: keypadread ( -- val )  \ read keypad and send back 0x0-0xF numbers
	CALL_KEYPAD xcall
	100 ms  \ 50ms works but you get multiple reports per press  \ CODE REVIEW: shorten this and see responsiveness
	1 read-com if rcv-buffer c@ else $FF then ;

: fastkeypad ( -- key )  \ software polling --- could do other things to interrupt/do other instructions
    begin 
        keypadread dup 
        $FF = if drop continue else exitloop then 
        key? if pollstop then pollstate@ or 
    until ;

: fastkeypad-chars ( -- )  \ emits which keys are pressed, as ASCII characters not numbers
	 cr pollcontinue begin fastkeypad pollstate@ if exitloop else bytes keypad-chars + c@ emit space continue then until ;

: fastkeypad-num ( -- n )  \ emits one number onto stack
	 cr pollcontinue fastkeypad pollstate@ if exitloop else bytes keypad-nums + c@ dup . space then ;

: fastkeypad-nums ( -- nums onto stack )  \ emits numbers pressed onto stack
	 cr pollcontinue begin fastkeypad pollstate@ if exitloop else bytes keypad-nums + c@ dup . space continue then until ;

: drum_action ( n -- ) 
    1 - dup 
    bytes keypad-nums + c@ . space  \ display
    [play-tone]  \ play-note
    ;
    
: fastkeypad-drums ( -- nums onto stack )  \ emits numbers pressed onto stack  WARNING: FOR PC GENERATION OF SOUND
        cr pollcontinue begin fastkeypad pollstate@ if exitloop else drum_action continue then until 
        ;
   
\ ---------------------------------------
\ EMBEDDED LOOP - MOST EFFICIENT

: play-drums ( -- nums onto stack )  \ emits numbers pressed onto stack  WARNING: FOR TARGET GENERATION OF SOUND
        cr CALL_LOOPMUSICKEYS xcall ." Keyboard sensing loop enabled.  Toggle interrupt switch to regain uC.";
   

\ ---------------------------------------

   
: drums-init 
   \ keypad-init7 
   tones-twohanddrums  ( tones-black-key-groove )  load-tones 
   drumroll
   7 setreg-tones
   20 13 setreg  \ tone duration
   80 14 setreg  \ delay before next tone (delay=Y-Xdur) -- largest acceptable is 80 ms, in line with c.100ms fast human repress
   ;

\ TODO: need to improve the responsiveness of the keys -- i.e. need to embed the response within the uC as well.  Use registers, like piano design

\ TODO: implement old-style phone letter selection... if second press within timeout, switch the letter, otherwise accept


\ ====== APPLICATION  (include 3iBeats.fs)
\  KEYPAD TEST -- include 3iKeypad.fs
 \  : beep ( -- ) buzzer hi 5 ms buzzer lo 50 ms ;
 \  : buzzcount ( N -- ) 0 do beep loop ;
 \  : keypad-demo ( -- ) 
 \      begin  fastkeypad 1 + buzzcount  key?  until key drop ;
        
\ APPLICATION - use 74165 to allow a simple ATTINY to monitor a keypad continuously and send out the results, so not to slow down the main uC.  Another application for the GA144...

 \  create beatspad 0 c, 1 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 
 \  : drawnote ( 1/0 -- ) dup 1 = if 14 else 22 then emit space ;
 \  : nibblespace? ( -- ) 1 + 4 mod 0 = if 2 spaces then ;  \ put space after each 4 bits (nibble) to aid visual perception
 \  : beats-demo ( -- )
 \  cr ." Enter a quartet of 16th-note beats using hex digits on keypad: " cr
 \      4 0 do fastkeypad bytes beatspad + c@ drawnote i nibblespace? loop >b4 1 Nseq-nibble-msb  \ get 8 entries, then loop it
 \  ; 

\ ======== KEYPAD PORTION =========
\ keypad-init
\  : beat-keypad ( -- )  \ input rhythm from 4x4 matrix keypad, 1 quarter note (4 beats) at a time.
\                     \ ex: 8 8 11 0   10 11 0 0   8 8 11 0   10 15 8 0
\      reset-stop resetbytes draw-menu
\      cr ." Input beats (hex digits)." cr
\      begin fastkeypad-num bytecount++ bytecount@ nseq-nibble-msb key? until key drop
\  ;
\ TODO: wrap into menu...
\ _________________________________


\ ====== APPLICATION \ ======== 5 button menu
( BUTTON LAYOUT:
    k3-play k2-clr k1-rest
     k4-clrall k5-beat
)

\ 5-button menu controls use Nano pins 2-6
\ 2 constant k1-rest
\ 3 constant k2-clr
\ 4 constant k3-play
\ 5 constant k4-clrall
\ 6 constant k5-beat
\ 
\ : buttons-init
\     k1-rest pullup \ buttons connect pin to ground and are active low
\     k2-clr pullup
\     k3-play pullup
\     k4-clrall pullup
\     k5-beat pullup
\ ;




\ =================================================

: help-drum
  cr
  cr ." DRUMS (help-drums) - 7-keys for 5 pc drum kit with hi-hat"
  cr 
  cr ." hardware: requires Nano, and 7-key buttons directly wired to nano pins 2-8"
  cr ." firmware: 3iDrumPads.avr which includes 3iForth.avr"
  cr
  cr ." include  3iForth.fs  3iSound-4x4.fs  3iBeats.fs  3iDrumPad.fs"
  cr ." 1 open-com     rx"
  cr
  cr ." setup-beatsboard   drums-init"  
  cr ." fastkeypad-nums"
  cr ." play-drums"
  cr
;

help-drum



\ FURTHER READING
\  The marvelous human auditory system: 
\     https://www.wnycstudios.org/podcasts/radiolab/articles/257870-ears-dont-lie




\ Versions
\ V0 - 2021-08-02---Mon--- - STARTED
\ V1 - 2021-08-03---Tue---07:55 - ADDED: drum tuning for 5 piece drum + hi-hat and alternate base
\ V2 - 2021-08-04---Wed---00:06 - FEATURE: Got attachInterrupt working with timing.  (Big deal - allows testing loops in Forth without having to recompile!)
\                   ADDED: documentation on interrupts - serial event is NOT an interrupt
\ V3 - 2021-08-04---Wed---06:50 - perfected both embedded musical keypad loop using attachInterrupt and forth controlled keypad loop using Forth state
